// --8<--8<--8<--8<--
//
// Copyright (C) 2006 Smithsonian Astrophysical Observatory
//
// This file is part of RDB
//
// RDB is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// RDB is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the 
//       Free Software Foundation, Inc. 
//       51 Franklin Street, Fifth Floor
//       Boston, MA  02110-1301, USA
//
// -->8-->8-->8-->8--

#include "RDB.h"

/**
 * \param is the input stream.
 * \param rdb the table to fill.
 *
 * \return A reference to the input stream.
 *
 * If the stream contains a valid rdbtable fill it... otherwise set
 * ios::failbit. 
 *
 */
istream&
operator>>( 
  istream& is, 
  RDB& rdb 
  ) throw ( RDBErr ) {

  try { 
    if ( &is != rdb._isptr ) {
      rdb.open( &is );
      
    } else {
      if ( RDB::REOF == rdb.read( ) ) {
	is.setstate( ios::eofbit | ios::failbit );

      }      
    }
  } catch ( RDBErr& rdberr ) {
    rdberr.set_message( "operator>>(istream&,RDB&):  " );
    throw( rdberr );

  } catch ( ... ) {
    throw( RDBErr( "operator>>(istream&,RDB&):  unexpected exception caught" ) );

  }

  return is;

}

/**
 * \param os the output stream.
 * \param rdb the table to print.
 *
 * \return A reference to the output stream
 *
 * Places the table on the output stream.
 *
 */
ostream& 
operator<<( 
  ostream& os, 
  RDB& rdb 
  ) throw ( RDBErr ) {

  try {
    if ( &os != rdb._osptr ) {
      rdb.open( &os );

    } else {
      rdb.write( );

    }
  } catch ( RDBErr& rdberr ) {
    rdberr.set_message( "operator<<(ostream&,RDB&):  " );
    throw( rdberr );
    
  } catch ( ... ) {
    throw( RDBErr( "operator<<(ostream&,RDB&):  unexpected exception caught" ) );
    
  }

  return os;

}

/**
 * \param name the name of the RDB file.
 * \param mode the ios::openmode of the file.
 *
 * \exception RDBErr error opening RDB file.
 * \exception RDBErr error parsing RDB comment or column name and definition.
 * \exception RDBErr error parsing RDB column definition.
 *
 * Attaches RDB object's I/O stream to the file.  If mode is set
 * to ios::in, it calls RDB#parseHeader(void) to read comments,
 * column names, and column definitions.
 *
 * If no filename is specified, then the RDB#open() method may be
 * used later to attach a file to the object.
 *
 */
RDB::RDB( 
  const string& name, 
  ios::openmode mode 
  ) throw ( RDBErr ) 
  : _filename(name), _mode(mode), 
    _isptr(NULL), _osptr(NULL),
    _myisptr(true), _myosptr(true),
    _rewindto(0), 
    _ncomms(0), _ncols(0), _nrows(0),
    _knowrows(false), _rownum(0),
    _autoidx(false),
    _firstread(true), _lastread(false), _writehdr(true),
    _comms(NULL), _cols(NULL),  
    _nrcol("_NR","N"), _mycols(NULL), 
    _line(1024,'\0') {
    
  _nrcol.mapData( &_rownum, 1 );
  
  if ( name.empty( ) ) {
    close( );

  } else {
    try {
      open( name, mode );
      
    } catch ( RDBErr& rdberr ) {
      rdberr.set_message( "RDB::RDB(string&,ios::openmode):  " );
      throw( rdberr );
      
    } catch ( ... ) { 
      throw( RDBErr( "RDB::RDB(string&,ios::openmode):  unexpected exception caught" ) );
      
    }
  }
}

/**
 * \param isptr input stream to attach to the RDB object.
 *
 * \exception RDBErr error parsing RDB comment or column name and definition. 
 * \exception RDBErr error parsing RDB column definition.
 *
 * Attaches the istream to the RDB object.  RDB#parseHeader(void) is
 * called to read comments, column names, and column definitions.
 *
 */
RDB::RDB( 
  istream* isptr 
  ) throw ( RDBErr ) : _filename("istream"), 
		       _isptr(NULL), _osptr(NULL),
		       _myisptr(false), _myosptr(false), 
		       _rewindto(0),
		       _ncomms(0), _ncols(0), _nrows(0),
		       _knowrows(false), _rownum(0), 
		       _autoidx(false),
		       _firstread(true), _lastread(false), _writehdr(true),
		       _comms(NULL), _cols(NULL),  
		       _nrcol("_NR","N"), _mycols(NULL),
		       _line(1024,'\0') {

  _nrcol.mapData( &_rownum, 1 );

  try {
    open( isptr );

  } catch ( RDBErr& rdberr ) {
    rdberr.set_message( "RDB::RDB(istream*):  " );
    throw( rdberr );
    
  } catch ( ... ) { 
    throw( RDBErr( "RDB::RDB(istream*):  unexpected exception caught" ) );
    
  }
} 

/**
 * \param osptr output stream to attach to the RDB object.
 *
 * Attaches the ostream to the RDB object.
 *
 */
RDB::RDB( 
  ostream* osptr 
  ) throw ( RDBErr ) : _filename("ostream"), 
		       _isptr(NULL), _osptr(NULL),
		       _myisptr(false), _myosptr(false),
		       _rewindto(0),
		       _ncomms(0), _ncols(0), _nrows(0),
		       _knowrows(false), _rownum(0),
		       _autoidx(false),
		       _firstread(true), _lastread(false), _writehdr(true),
		       _comms(NULL), _cols(NULL),  
		       _nrcol("_NR","N"), _mycols(NULL),
		       _line(1024,'\0') {   

  _nrcol.mapData( &_rownum, 1 );

  try {
    open( osptr );

  } catch ( RDBErr& rdberr ) {
    rdberr.set_message( "RDB::RDB(ostream&):  " );
    throw( rdberr );
    
  } catch ( ... ) { 
    throw( RDBErr( "RDB::RDB(ostream&):  unexpected exception caught" ) );
    
  }
} 

/**
 * \param rdb copy RDB object.
 *
 * Makes a copy of the argument. 
 *
 */
RDB::RDB( 
  const RDB& rdb
  ) throw ( RDBErr ) : _filename(rdb._filename), _mode(rdb._mode), 
		       _isptr(NULL), _osptr(NULL), 
		       _myisptr(true), _myosptr(true),
		       _rewindto(0), _ncomms(0), _ncols(0), _nrows(0),
		       _knowrows(false), _rownum(0),
		       _autoidx(false), _firstread(true), _lastread(false),
		       _writehdr(true),
		       _comms(NULL), _cols(NULL), _nrcol("_NR","N"),
		       _mycols(NULL), _line(1024,'\0') { 
    
  _nrcol.mapData( &_rownum, 1 );

  try { 
    open( rdb );

  } catch ( RDBErr& rdberr ) {
    rdberr.set_message( "RDB::RDB(RDB&):  " );
    throw( rdberr );
    
  } catch ( ... ) { 
    throw( RDBErr( "RDB::RDB(RDB&):  unexpected exception caught" ) );
    
  }
} 

/**
 * Deletes RDBComments and RDBColumns.  Closes only the streams the
 * object opened.  User is responsible for closing streams they supply
 * to the object.
 *
 * \warning If the user supplies an RDBColumn object via the
 * RDB#setColumn() method, the user is responsible for deleting that
 * object. 
 *
 */
RDB::~RDB(
  void
  ) {

  close( );

  if ( _comms ) {
    delete [] _comms; 

  }

  if ( _cols ) {
    for ( size_t idx = 0; idx < _ncols; idx++ ) {
      if ( _mycols[idx] ) {
	delete _cols[idx]; 
      }
    }
    delete [] _cols;
    delete [] _mycols;
  }

  _isptr  = NULL;
  _osptr  = NULL;
  _comms  = NULL;
  _cols   = NULL;
  _mycols = NULL;

}

/**
 * \param name the name of the RDB file.
 * \param mode the ios::openmode of the file.
 *
 * \exception RDBErr error opening RDB file.
 * \exception RDBErr error parsing RDB comment or column name and definition. 
 * \exception RDBErr error parsing RDB column definition.
 *
 * Attaches RDB object's I/O stream to the file.  If mode is set
 * to ios::in, it calls RDB#parseHeader(void) to read comments,
 * column names, and column definitions.
 *
 * \warning This method closes any previously opened streams.
 *
 */
void
RDB::open( 
  const string& name, 
  ios::openmode mode 
  ) throw ( RDBErr ) { 

  close( );

  _filename = name;
  _mode     = mode;

  if ( ios::in & _mode ) {
    _isptr = new ifstream( _filename.c_str( ), _mode );
    _myisptr = true;
    if ( ! _isptr || ! *_isptr ) {
      stringstream ss;
      ss << "RDB::open(string&,ios_base::openmode):  Error opening '" 
	 << _filename << "' with mode";
      if ( ios_base::app & _mode ) 
	ss << " ios_base::app";
      if ( ios_base::binary & _mode )
	ss << " ios_base::binary";
      if ( ios_base::in & _mode )
	ss << " ios_bae::in";
      if ( ios_base::out & _mode )
	ss << " ios_base::out";
      if ( ios_base::trunc & _mode )
	ss << " ios_base::trunc";
      if ( ios_base::ate & _mode )
	ss << " ios_base::ate";
      throw( RDBErr( ss.str( ) ) );

    }

    try { 
      parseHeader( );

    } catch ( RDBErr& rdberr ) {
      rdberr.set_message( "RDB::open(string&,ios::openmode):  error with file '" + _filename + "':  " );
      throw( rdberr );
      
    } catch ( ... ) { 
      throw( RDBErr( "RDB::open(string&,ios::openmode):  unexpected exception caught with file '" + _filename + "'" ) );
      
    }
  }

  if ( ios::out & _mode ) {
    _osptr = new ofstream( _filename.c_str( ), _mode );
    _myosptr = true;
    if ( ! _osptr || ! *_osptr ) {
      stringstream ss;
      ss << "RDB::open(string&,ios_base::openmode):  Error opening '" 
	 << _filename << "' with mode";
      if ( ios_base::app & _mode ) 
	ss << " ios_base::app";
      if ( ios_base::binary & _mode )
	ss << " ios_base::binary";
      if ( ios_base::in & _mode )
	ss << " ios_bae::in";
      if ( ios_base::out & _mode )
	ss << " ios_base::out";
      if ( ios_base::trunc & _mode )
	ss << " ios_base::trunc";
      if ( ios_base::ate & _mode )
	ss << " ios_base::ate";
      throw( RDBErr( ss.str( ) ) );

    }
  }
}

/**
 * \param isptr input stream to attach to the RDB object.
 *
 * \exception RDBErr error parsing RDB comments or column names and definitions 
 * \exception RDBErr error parsing RDB column definition.
 *
 * Attaches the istream to the RDB object.  RDB#parseHeader(void) is
 * called to read comments, column names, and column definitions.
 *
 * \warning This method closes any previously opened streams.
 *
 */
void
RDB::open( 
  istream* isptr
  ) throw ( RDBErr ) { 

  close( );

  _filename = "istream";
  _mode     = ios::in;
  _isptr    = isptr;

  try { 
    parseHeader( );

  } catch ( RDBErr& rdberr ) {
    rdberr.set_message( "RDB::open(istream&):  " );
    throw( rdberr );
    
  } catch ( ... ) { 
    throw( RDBErr( "RDB::open(istream&):  unexpected exception caught" ) );
    
  }
}

/**
 * \param osptr output stream to attach to the RDB object.
 *
 * Attaches the ostream to the RDB object.
 *
 * \warning This method closes any previously opened streams.
 *
 */
void
RDB::open( 
  ostream* osptr
  ) throw ( RDBErr ) { 

  close( );

  _filename = "ostream";
  _mode     = ios::out;
  _osptr    = osptr;

}

/**
 * \param rdb copy RDB object.
 *
 * \warning This method is currently not implemented.
 *
 */
void
RDB::open( 
  const RDB& rdb
  ) throw ( RDBErr ) { 
  
  close( );

  _filename = rdb._filename;
  _mode     = rdb._mode;

  if ( ios::in & _mode ) {
    _isptr = new ifstream( _filename.c_str( ), _mode );
    _myisptr = true;
    if ( ! _isptr ) {
      stringstream ss;
      ss << "RDB::open(RDB&):  Error opening '" 
	 << _filename << "' with mode";
      if ( ios_base::app & _mode ) 
	ss << " ios_base::app";
      if ( ios_base::binary & _mode )
	ss << " ios_base::binary";
      if ( ios_base::in & _mode )
	ss << " ios_bae::in";
      if ( ios_base::out & _mode )
	ss << " ios_base::out";
      if ( ios_base::trunc & _mode )
	ss << " ios_base::trunc";
      if ( ios_base::ate & _mode )
	ss << " ios_base::ate";
      throw( RDBErr( ss.str( ) ) );

    }

    try { 
      parseHeader( );

    } catch ( RDBErr& rdberr ) {
      rdberr.set_message( "RDB::open(RDB&):  error with file '" + _filename + "':  " );
      throw( rdberr );
      
    } catch ( ... ) { 
      throw( RDBErr( "RDB::open(RDB&):  unexpected exception caught with file '" + _filename + "'" ) );
      
    }
  }

  if ( ios::out & _mode ) {
    _osptr = new ofstream( _filename.c_str( ), _mode );
    _myosptr = true;
    if ( ! _osptr ) {
      stringstream ss;
      ss << "RDB::open(RDB&):  Error opening '" 
	 << _filename << "' with mode";
      if ( ios_base::app & _mode ) 
	ss << " ios_base::app";
      if ( ios_base::binary & _mode )
	ss << " ios_base::binary";
      if ( ios_base::in & _mode )
	ss << " ios_bae::in";
      if ( ios_base::out & _mode )
	ss << " ios_base::out";
      if ( ios_base::trunc & _mode )
	ss << " ios_base::trunc";
      if ( ios_base::ate & _mode )
	ss << " ios_base::ate";
      throw( RDBErr( ss.str( ) ) );

    }
  }
}

/**
 * Closes any streams that the RDB object created.
 *
 */
void
RDB::close(
  void
  ) { 

  if ( _myisptr && _isptr ) {
    delete _isptr;

  }

  if ( _osptr && _writehdr ) {
    size_t idx;
    for ( idx = 0; idx < _ncomms; idx++ )
      *_osptr << _comms[idx] << endl;

    if ( _ncols ) {
      for ( idx = 0; idx < _ncols - 1; idx++ )
	*_osptr << _cols[idx]->getName( ) << "\t";
      *_osptr << _cols[idx]->getName( ) << endl;
      
      for ( idx = 0; idx < _ncols - 1; idx++ )
	*_osptr << _cols[idx]->getDef( ) << "\t";
      *_osptr << _cols[idx]->getDef( ) << endl;

      _rewindto = _osptr->tellp( );
      _rownum   = 0;
      _writehdr = false;
    
    }
  }

  if ( _myosptr && _osptr ) {
    delete _osptr;

  }

  _filename.erase( );
  _isptr    = NULL;
  _myisptr  = false;
  _osptr    = NULL;
  _myosptr  = false;
  _autoidx  = false;
  _nrows    = 0;
  _knowrows = false;
  _rownum   = 0;

}

/**  
 * \exception RDBErr error if the number of tokens found does not match the number of tokens expected.
 * \exception RDBErr error if non-numeric data is found in a numeric column.
 * \exception RDBErr error if a floating point number is being assigned to an integer column.
 *
S * \return RDB::Status indicating if end of file, end of line, or end
 * of group was reached.
 *
 * Reads, parses, and stores data from the next line in the RDB file
 * is an input stream has been provided.  If RDB#_autoidx is set, this
 * method will advance the indices in each RDBColumn's data pointer
 * before reading and will increment the row number, RDB#_rownum, for
 * the default `_NR' column. 
 *
 */
int
RDB::read( 
  void
  ) throw ( RDBErr ) { 

  if ( ios::in & _mode && _isptr && *_isptr ) {
    int newgroup = RDB::REOF;
    char c = _isptr->peek( );
    if ( '\n' == c ) {
      _isptr->get( );
      c = _isptr->peek( );

    }

    if ( EOF == c ) {
      if ( _lastread )
	return RDB::REOF;
      else 
	_lastread = true;
    }

    _rownum++;

    if ( ! _lastread ) {
      try {
	getline( *_isptr, _line, '\n' );
	_isptr->clear( );
	vector<string> tokens;
	suplib::tok( tokens, _line, "\t", false );

	if ( tokens.size( ) != _ncols ) {
	  stringstream ss;
	  ss << "RDB::read(void):  error reading file '" << _filename << "':  invalid row at " << _rownum << ":  number of token(" << tokens.size( ) << ") != number of columns(" << _ncols << ")";
	  throw( RDBErr( ss.str( ) ) );
	}

	for ( int idx = 0; idx < _ncols; idx++ ) {
	  _cols[idx]->setData( tokens[idx] );
	  if ( _autoidx ) {
	    _cols[idx]->advanceIdx( );
	  }
	  newgroup = newgroup | _cols[idx]->newGroup( );

	}
      } catch ( RDBErr& rdberr ) {
	rdberr.set_message( "RDB::read(void):  error reading file '" + _filename + "':  " );
	throw( rdberr );
	
      } catch ( ... ) { 
	throw( RDBErr( "RDB::read(void):  unexpected exception caught with file '" + _filename + "'" ) );
	
      }

      if ( _firstread && (RDB::REOL ^ newgroup) ) {
	try {
	  getline( *_isptr, _line, '\n' );
	  if ( _isptr->eof( ) ) {
	    newgroup |= REOG;
	    for ( int idx = 0; idx < _ncols; idx++ ) {
	      if ( _autoidx ) {
		_cols[idx]->advanceIdx( );
	      }
	    }
	  } else {
	    _isptr->clear( );
	    vector<string> tokens;
	    suplib::tok( tokens, _line, "\t", false );
	  
	    if ( tokens.size( ) != _ncols ) {
	      stringstream ss;
	      ss << "RDB::read(void):  error reading file '" << _filename << "':  invalid row at " << _rownum+1 << ":  number of token(" << tokens.size( ) << ") != number of columns(" << _ncols << ")";
	      throw( RDBErr( ss.str( ) ) );
	    }

	    for ( int idx = 0; idx < _ncols; idx++ ) {
	      _cols[idx]->setData( tokens[idx] );
	      if ( _autoidx ) {
		_cols[idx]->advanceIdx( );
	      }
	      //	    newgroup = _cols[idx]->newGroup( );
	    }
	  }
	} catch ( RDBErr& rdberr ) {
	  rdberr.set_message( "RDB::read(void):  error reading file '" + _filename + "':  " );
	  throw( rdberr );
	  
	} catch ( ... ) { 
	  throw( RDBErr( "RDB::read(void):  unexpected exception caught reading file '" + _filename + "'" ) );
	  
	}

	_firstread = false;
	
      }
    } else {
      if ( _autoidx ) {
	advanceIdx( );
	newgroup = RDB::REOG;
      } else {
	newgroup = RDB::REOF;
      }
    }
    return newgroup;

  } else {
    return RDB::REOF;

  }
}

/**
 * \return false if EOF or if the ostream is bad.
 *
 * Writes the next line of data to the RDB file, if an output stream
 * has been provided.  If the RDB header has not been written yet, it
 * outputs the comments, column names and column definitions.  If
 * RDB#_autoidx is set, it will advance the indices in the RDBColumn's
 * data pointer before reading and will increment the row number,
 * RDB#_rownum, for the default `_NR' column. 
 * 
 * After a successful write, the RDB#_autoidx flag is set to true. 
 *
 */
bool
RDB::write( 
  void
  ) throw ( RDBErr ) { 

  if ( ios::out & _mode && _osptr ) {
    if ( _writehdr ) {
      size_t idx;
      for ( idx = 0; idx < _ncomms; idx++ )
	*_osptr << _comms[idx] << endl;
      
      for ( idx = 0; idx < _ncols - 1; idx++ )
	*_osptr << _cols[idx]->getName( ) << "\t";
      *_osptr << _cols[idx]->getName( ) << endl;
      
      for ( idx = 0; idx < _ncols - 1; idx++ )
	*_osptr << _cols[idx]->getDef( ) << "\t";
      *_osptr << _cols[idx]->getDef( ) << endl;
      
      _rewindto = _osptr->tellp( );
      _rownum   = 0;
      _writehdr = false;
      
    }

    if ( _autoidx ) {
      advanceIdx( );

    }

    for ( size_t idx = 0; idx < _ncols; idx++ ) {
      *_osptr << *_cols[idx] << (( idx < _ncols - 1 ) ? ( '\t' ) : ( '\n' ));

    }

    _rownum++;

    return true;

  } else {
    return false;

  }
}

/**
 * \return false if the stream associated with this RDB object cannot
 * be rewound, i.e seekg() or seekp() is not defined.
 *
 * Rewinds the stream associate with this object to the beginning of
 * the first row of data.  It also calls RDBColumn#rewind(void) for
 * each RDBColumn.  If the user supplied data storage to a particular
 * column, the auto-incrementing index is rewound to point at the
 * first element of the user supplied array. 
 *
 * \warning If an istream or ostream, like cin or cout, was provided
 * to the object, you only get one chance to read it.  You can't
 * rewind, you can't close and reopen.
 *
 */
bool
RDB::rewind( 
  void
  ) throw ( RDBErr ) { 

  if ( ios::in & _mode ) {
    _isptr->clear( );
    _isptr->seekg( _rewindto );
    _rownum = 0;
    
    if ( _rewindto != (size_t)_isptr->tellg( ) ) {
      return false;

    } else {
      for ( size_t idx = 0; idx < _ncols; idx++ ) {
	_cols[idx]->rewind( );

      }

      return true;

    }
  } else if ( ios::out & _mode ) {
    _osptr->clear( );
    _osptr->seekp( _rewindto );
    _rownum = 0;
    
    if ( _rewindto != (size_t)_osptr->tellp( ) ) {
      return false;

    } else {
      for ( size_t idx = 0; idx < _ncols; idx++ ) {
	_cols[idx]->rewind( );

      }

      return true;

    }
  }

  return false;

}

/**
 * \param on sets the auto-indexing behavior for the RDB object and
 * its columns.
 *
 * Sets RDBTable#_autoidx.  This effectly turns off the
 * auto-incrementing behavior for a single call to RDB#read(void)
 * or RDB#write(void).  For arrays of mapped memory, this will
 * cause one line's data to overwrite the previous line's data.
 *
 * \warning RDB#_autoidx is set to true after each call to
 * RDB#read(void) or RDB#write(void). 
 *
 */
void
RDB::autoIdx(
  const bool on
  ) { 
  
  _autoidx = on;

}

/**
 * \return State of the auto-indexing flag for this object.
 *
 */
bool
RDB::autoIdx( 
  void
  ) const { 

  return _autoidx;

}

/**
 * This method advances the auto-indices for all RDBColumns associated
 * with this object.
 *
 */
void
RDB::advanceIdx( 
  void
  ) {

  for ( size_t idx = 0; idx < _ncols; idx++ ) {
    _cols[idx]->advanceIdx( );

  }
  _nrcol.advanceIdx( );

}

/**
 * \param group a bool indicating whether group is on or off.
 * \param name the name of the column in this RDB object.
 *
 * \exception RDBEr error if there is no column with matching
 * name.
 *
 */
void
RDB::setGroup( 
  const string& name, 
  bool group
  ) throw ( RDBErr ) {

  _autoidx = true;

  for ( size_t jdx = 0; jdx < _ncols; jdx++ ) {
    if ( name == _cols[jdx]->getName( ) ) {
      _cols[jdx]->setGroup( group );
      return;

    }
  }  

  
  if ( "_NR" != name ) {
    throw( RDBErr( string("RDB::setGroup(string):  error in file '" + _filename + "':  Column ") + name + "(0) not found" ) );

  }

  _nrcol.setGroup( group );
}

/**
 * \param group a bool indicating whether group is on or off.
 * \param idx the position of the column on which to operate.
 *
 * \exception RDBErr error if there is no column with matching
 * name.
 *
 */
void
RDB::setGroup( 
  const int idx, 
  bool group 
  ) throw ( RDBErr ) {
  
  _autoidx = true;

  if ( 0 <= idx && idx < (int)_ncols ) {
    _cols[idx]->setGroup( group );

  } else {    
    stringstream ss;
    ss << "RDB::setGroup(string):  error with file '" + _filename + "':  Column "  << idx << " of " << _ncols << " not found";
    throw( RDBErr( ss.str( ) ) );

  }
}

/**
 * \param name the name of the column.
 *
 * \exception RDBErrNotFnd error if there is no column with matching name.
 *
 * \return Status of the column's group flag.
 *
 */
bool
RDB::getGroup( 
  const string& name
  ) throw ( RDBErr ) {

  for ( size_t jdx = 0; jdx < _ncols; jdx++ ) {
    if ( name == _cols[jdx]->getName( ) ) {
      return _cols[jdx]->getGroup( );

    }
  }

  if ( "_NR" != name ) {
    throw( RDBErr( string("RDB::getGroup(string):  error with file '" + _filename + "':  Column ") + name + "(0) not found" ) );

  }

  return _nrcol.getGroup( );

}

/**
 * \param idx the column index.
 *
 * \exception RDBErr error if the index is out of range.
 *
 * \return Status of the column's group flag.
 *
 */
bool 
RDB::getGroup( 
  const int idx
  ) throw ( RDBErr ) {

  if ( 0 > idx && idx >= _ncols ) {
    stringstream ss;
    ss << "RDB::getGroup(string):  error with file '" + _filename + "':  Column " << idx << " out of " << _ncols << " not found";
    throw( RDBErr( ss.str( ) ) );
    
  }

  return _cols[idx]->getGroup( );
}

/**
 * \return True if any column indicates a new group, false otherwise.
 *
 * A new group occurs when the data values between consecutive rows
 * change.
 *
 */
bool
RDB::newGroup( 
  void
  ) {
  
  for ( size_t idx = 0; idx < _ncols; idx++ ) {
    if ( REOL ^ _cols[idx]->newGroup( ) ) {
      return true;

    }
  }

  return false;

}
  
/**
 * \param comm a comment string.
 * \param idx the position of the comment on which to operate.
 *
 * Comments are numbered as they appear in the RDB file from top to
 * bottom.  If the index is not specified or does not fall within the
 * range of existing comments, the RDBComment is appended to the list
 * of comments.  
 *
 */
void 
RDB::setComment( 
  const string& comm,
  const int idx 
  ) { 

  if ( 0 < idx && idx < (int) _ncomms ) {
    _comms[idx] = comm;

  } else {
    RDBComment* tmp = new RDBComment[_ncomms+1];

    if ( _comms ) {
      for ( size_t jdx = 0; jdx < _ncomms; jdx++ ) {
	tmp[jdx] = _comms[jdx];

      }
      delete [] _comms;

    }

    tmp[_ncomms++] = comm;

    _comms         = tmp;

  }
}

/**
 * \param comm a reference to the RDBComment object
 * \param idx the position of the comment on which to operate.
 *
 * Comments are numbered as they appear in the RDB file from top to
 * bottom.  If the index is not specified or does not fall within the
 * range of existing comments, the RDBComment is appended to the list
 * of comments.  
 *
 */
void 
RDB::setComment( 
  RDBComment& comm,
  const int idx 
  ) { 

  if ( 0 < idx && idx < (int) _ncomms ) {
    _comms[idx] = comm;

  } else {
    RDBComment* tmp = new RDBComment[_ncomms+1];

    if ( _comms ) {
      for ( size_t jdx = 0; jdx < _ncomms; jdx++ ) {
	tmp[jdx] = _comms[jdx];

      }
      delete [] _comms;

    }

    tmp[_ncomms++] = comm;

    _comms         = tmp;

  }
}

/**
 * \param comm a reference to the RDBComment object
 * \param name the name of the comment keyword.
 * \param idx currently not defined.
 *
 * If no matching comment keyword is found, the comment is appended to
 * the list of comments.  Otherwise, comm replaces the existing comment.
 *
 */
void
RDB::setComment( 
  RDBComment& comm, 
  const string& name, 
  const size_t idx 
  ) { 

  for ( size_t jdx = 0; jdx < _ncomms; jdx++ ) {
    if ( name == _comms[jdx].getKeyword( ) ) {
      _comms[jdx] = comm;
      return;

    }
  }

  setComment( comm, -1 );

}

/**
 * \param rdb a reference to an RDB object
 *
 * Copies all comments from the argument to this RDB object.  Any
 * comments previously associated with this object are removed.
 *
 */
void
RDB::setComment( 
  const RDB& rdb
  ) { 

  if ( rdb._ncomms > _ncomms ) {
    if ( _ncomms )
      delete [] _comms;
    _comms = new RDBComment[rdb._ncomms];

  }

  _ncomms = rdb._ncomms;

  for ( size_t idx = 0; idx < _ncomms; idx++ ) {
    _comms[idx] = rdb._comms[idx];

  }
}

/**
 * \param idx the index of the RDBComment.
 *
 * \exception RDBErrNot error if the index is out of range.
 *
 * \return RDBComment with index equal to idx.
 *
 * RDBComments are indexed in the order in which they appear in the
 * RDB file starting with 0.  Parameter idx must be between 0 and
 * RDB::nComments(void) less one.
 *
 */
RDBComment&
RDB::getComment( 
  const size_t idx
  ) throw ( RDBErr ) { 

  if ( 0 > idx && idx >= _ncomms ) {
    stringstream ss;
    ss << "RDB::getComment(size_t):  error with file '" + _filename + "':  Comment "  << idx << "out of "<< _ncomms << " not found";
    throw( RDBErr( ss.str( ) ) );

  }

  return _comms[idx];

}

/**
 * \param name the name of the comment keyword.
 * \param idx currently not defined.
 *
 * \exception RDBErr error if there is no comment with matching keyword.
 *
 * \return RDBComment with matching keyword.
 *
 */
RDBComment&
RDB::getComment( 
  const string& name, 
  const size_t idx
  ) throw ( RDBErr ) { 

  for ( size_t jdx = 0; jdx < _ncomms; jdx++ ) {
    if ( name == _comms[jdx].getKeyword( ) ) {
      return _comms[jdx];

    }
  }

  stringstream ss;
  ss << "RDB::getComment(string):  error with file '" << _filename<< "':  Comment "  << name << "("<< idx << ") not found";
  throw( RDBErr( ss.str( ) ) );

}

/**
 * \param name of the RDBColumn object
 * \param def definition of the RDBColumn object
 * \param idx the position of the column on which to operate.
 *
 * Columns are numbered as they appear in the RDB file from left to
 * right starting with 0.  If the index is not specified or does not
 * fall within the range of existing columns, an RDBColumn is created
 * and is appended to the list of columns.  The RDB object handles
 * memory allocation and deletion.
 *
 * Upon completion, the RDB object contains a pointer to the RDBColumn
 * created.  Any modifications to the RDBColumn outside of the
 * RDB object will be reflected within the RDB object and vice versa.
 *
 */
void
RDB::setColumn( 
  const string& name, 
  const string& def,
  const int idx 
  ) { 

  if ( 0 <= idx && idx < (int)_ncols ) {

    if ( _mycols[idx] ) {
      delete _cols[idx];

    }

    _cols[idx]->setName( name );
    _cols[idx]->setDef( def );
    _mycols[idx] = true;

  } else {
    RDBColumn** tmp = new RDBColumn*[_ncols+2];
    bool* tmpmycols = new bool[_ncols+1];
    RDBColumn* col;
    int pos;
    if ( def.npos != (pos = def.find_first_not_of( "0123456789" )) && 'N' == def[pos] ) {
      col = new RDBDoubleColumn( name, def );
    } else { 
      col = new RDBStringColumn( name, def );
    }

    if ( _cols ) {
      for ( size_t jdx = 0; jdx < _ncols; jdx++ ) {
	tmp[jdx]       = _cols[jdx];
	tmpmycols[jdx] = _mycols[jdx];

      }
      
      delete [] _cols;
      delete [] _mycols;

    }

    tmp[_ncols] = col;
    tmpmycols[_ncols++] = true;

    _cols   = tmp;
    _mycols = tmpmycols;

  }
}

/**
 * \param col a reference to the RDBColumn object
 * \param idx the position of the column on which to operate.
 *
 * Columns are numbered as they appear in the RDB file from left to
 * right starting with 0.  If the index is not specified or does not
 * fall within the range of existing columns, the RDBColumn is
 * appended to the list of columns.
 *
 * Upon completion, the RDB object contains a pointer to the RDBColumn
 * provided.  Any modifications to the RDBColumn outside of the
 * RDB object will be reflected within the RDB object and vice versa.
 *
 * \warning The user is responsible for freeing the RDBColumn. 
 *
 */
void
RDB::setColumn( 
  RDBColumn* col,
  const int idx 
  ) { 

  if ( 0 <= idx && idx < (int)_ncols ) {

    if ( _mycols[idx] ) {
      delete _cols[idx];

    }

    _cols[idx]   = col;
    _mycols[idx] = false;

  } else {
    RDBColumn** tmp  = new RDBColumn*[_ncols+2];
    bool* tmpmycols  = new bool[_ncols+1];

    if ( _cols ) {
      for ( size_t jdx = 0; jdx < _ncols; jdx++ ) {
	tmp[jdx]       = _cols[jdx];
	tmpmycols[jdx] = _mycols[jdx];

      }
      
      delete [] _cols;
      delete [] _mycols;

    }

    tmp[_ncols]         = col;
    tmpmycols[_ncols++] = false;

    _cols   = tmp;
    _mycols = tmpmycols;

  }
}

/**
 * \param col a reference to the RDBColumn object
 * \param name the name of the column in this RDB object.
 * \param idx currently not defined.
 *
 * If no matching column is found, the column is appended to
 * the list of columns after assigning it a name of 'name'.
 * Otherwise, col replaces the existing column. 
 *
 * Upon completion, the RDB object contains a pointer to the RDBColumn
 * provided.  Any modifications to the RDBColumn outside of the
 * RDB object will be reflected within the RDB object and vice versa.
 *
 * \warning The user is responsible for freeing the RDBColumn. 
 *
 */
void 
RDB::setColumn( 
  RDBColumn* col, 
  const string& name, 
  const size_t idx 
  ) { 

  for ( size_t jdx = 0; jdx < _ncols; jdx++ ) {
    if ( name == _cols[jdx]->getName( ) ) {
      if ( _mycols[jdx] ) {
	delete _cols[jdx];

      }

      _cols[jdx]   = col;
      _mycols[jdx] = false;
      return;

    }
  }

  col->setName( name );
  setColumn( col, -1 );

}

/**
 * \param rdb a reference to an RDB object
 *
 * Copies all columns from the argument to this RDB object.  Any
 * columns previously associated with this object are removed.
 *
 * \warning The user is responsible for free any RDBColumns previously
 * allocated and assigned to this RDB object via the RDB#setColumn
 * methods.
 *
 */
void 
RDB::setColumn( 
  const RDB& rdb
  ) { 

  size_t idx = 0;
  if ( rdb._ncols > _ncols ) {
    RDBColumn** tmp = new RDBColumn*[rdb._ncols+1];
    bool* tmpmycols = new bool[rdb._ncols];

    if ( _cols ) {
      for ( idx = 0; idx < _ncols; idx++ ) {
	tmp[idx]       = rdb._cols[idx];
	tmpmycols[idx] = false;

	if ( _mycols[idx] ) {
	  delete _cols[idx];

	}
      }
      
      if ( _ncols ) {
	delete [] _cols;
	delete [] _mycols;

      }
    }
    
    for ( ; idx < rdb._ncols; idx++ ) {
      tmp[idx]       = rdb._cols[idx];
      tmpmycols[idx] = false;
      
    }
    _cols   = tmp;
    _mycols = tmpmycols;

  } else {
    for ( idx = 0; idx < rdb._ncols; idx++ ) {
      _cols[idx]   = rdb._cols[idx];
      _mycols[idx] = false;

    }

    for ( ; idx < _ncols; idx++ ) {
      delete _cols[idx];

    }
  }

  _ncols = rdb._ncols;

}

/**
 * \param idx the index of the RDBColumn.
 *
 * \exception RDBErr error if the index is out of range.
 *
 * \return RDBColumn with index equal to idx.
 *
 * RDBColumns are indexed in the order in which they appear in the
 * RDB file starting with 0.  Parameter idx must be between 0 and
 * RDB::nColumns(void) less one.
 *
 */
RDBColumn*
RDB::getColumn(
  const size_t idx
  ) throw ( RDBErr ) { 

  if ( 0 > idx && idx >= _ncols ) {
    stringstream ss;
    ss << "RDB::getColumn(size_t):  error with file '" << _filename << "':  Column "  << idx << " out of "<< _ncols << " not found";
    throw( RDBErr( ss.str( ) ) );
    
  }

  return _cols[idx];

}

/**
 * \param name the name of the column.
 * \param idx currently not defined.
 *
 * \exception RDBErrNotFnd error if there is no column with matching name.
 *
 * \return RDBColumn with matching name.
 *
 */
RDBColumn* 
RDB::getColumn( 
  const string& name,
  const size_t idx 
  ) throw ( RDBErr ) {

  for ( size_t jdx = 0; jdx < _ncols; jdx++ ) {
    if ( name == _cols[jdx]->getName( ) ) {
      return _cols[jdx];

    }
  }

  if ( "_NR" != name ) {
    stringstream ss;
    ss << "RDB::getColumn(string):  error with file '" << _filename << "':  Column "  << name << "("<< idx << ") not found";
    throw( RDBErr( ss.str( ) ) );

  }

  return &_nrcol;

}

/**
 * \param idx index of RDBColumn.
 * \param name to assign to RDBColumn at idx.
 *
 * \exception RDBErr if index is out of range.
 *
 */
void RDB::setName( const size_t idx, const string& name ) throw ( RDBErr ) {

  if ( idx >= (int)_ncols ) {
    stringstream ss;
    ss << "RDB::setName:  error with file '" << _filename << "':  index out of range [0," << idx << ")";
    throw( RDBErr( ss.str( ) ) );

  }
 
  _cols[idx]->setName( name );
  
}

/**
 * \param idx index of RDBColumn. 
 * \param def RDBColumn definition.
 *
 * \exception RDBErr if index is out of range.
 * \exception RDBErr if def is invalid RDBColumn definition.
 *
 */
void RDB::setDef( const size_t idx, const string& def  ) throw ( RDBErr ) {

  if ( idx > (int)_ncols ) {
    stringstream ss;
    ss << "RDB::setDef:  error with file '" << _filename << "':  index out of range [0," << idx << ")";
    throw( RDBErr( ss.str( ) ) );
  }

  try {
    _cols[idx]->setDef( def );

  } catch ( RDBErr& e ) {
    e.set_message( "RDB::setDef:  error with file '" + _filename + "':  " );
    throw( e );

  } catch ( ... ) {
    throw( RDBErr( "RDB::setDef:  unexpected exception caught with file '" + _filename +"'" ) );
  }
}

/**
 * \param idx index of RDBColumn.
 * \param width RDBColumn width.
 *
 * \exception RDBErr if index is out of range.
 * \exception RDBErr if width is invalid RDBColumn width.
 *
 */
void RDB::setWidth( const size_t idx, const long width ) throw ( RDBErr ) {

  if ( idx > (int)_ncols ) {
    stringstream ss;
    ss << "RDB::setWidth:  error with file '" << _filename << "':  index out of range [0," << idx << ")";
    throw( RDBErr( ss.str( ) ) );
  }

  try {
    _cols[idx]->setWidth( width );

  } catch ( RDBErr& e ) {
    e.set_message( "RDB::setWidth:  error with file '" + _filename + "':  " );
    throw( e );

  } catch ( ... ) {
    throw( RDBErr( "RDB::setWidth:  unexpected exception caught with file '" + _filename + "'" ) );
  }
}

/**
 * \param idx index of RDBColumn.
 * \param type RDBColumn type.
 *
 * \exception RDBErr if index is out of range.
 * \exception RDBErr if type is invalid RDBColumn type.
 *
 */
void RDB::setType( const size_t idx, const RDBColumn::Type type ) throw ( RDBErr ) {

  if ( idx > (int)_ncols ) {
    stringstream ss;
    ss << "RDB::setType:  error with file '" << _filename << "':  index out of range [0," << idx << ")";
    throw( RDBErr( ss.str( ) ) );
  }

  try {
    _cols[idx]->setType( type );

  } catch ( RDBErr& e ) {
    e.set_message( "RDB::setType:  error with file '" + _filename + "':  " );
    throw( e );

  } catch ( ... ) {
    throw( RDBErr( "RDB::setType:  unexpected exception caught with file '" + _filename + "'" ) );
  }
}

/**
 * \param idx index of RDBColumn.
 * \param just RDBColumn justification.
 *
 * \exception RDBErr if index is out of range.
 * \exception RDBErr if just is invalid RDBColumn justification.
 *
 */
void RDB::setJust( const size_t idx, const RDBColumn::Just just ) throw ( RDBErr ) {

  if ( idx > (int)_ncols ) {
    stringstream ss;
    ss << "RDB::setJust:  error with file '" << _filename << "':  index out of range [0," << idx << ")";
    throw( RDBErr( ss.str( ) ) );
  }

  try {
    _cols[idx]->setJust( just );

  } catch ( RDBErr& e ) {
    e.set_message( "RDB::setJust:  error with file '" + _filename + "':  " );
    throw( e );

  } catch ( ... ) {
    throw( RDBErr( "RDB::setJust:  unexpected exception caught with file '" + _filename + "'" ) );
  }
}

/**
 * \param idx index of RDBColumn. 
 * \param desc RDBColumn description.
 *
 * \exception RDBErr if index is out of range.
 * \exception RDBErr if desc is invalid RDBColumn description.
 *
 */
void RDB::setDesc( const size_t idx, const string& desc ) throw ( RDBErr ) {

  if ( idx > (int)_ncols ) {
    stringstream ss;
    ss << "RDB::setDesc:  error with file '" << _filename << "':  index out of range [0," << idx << ")";
    throw( RDBErr( ss.str( ) ) );
  }

  try {
    _cols[idx]->setDesc( desc );

  } catch ( RDBErr& e ) {
    e.set_message( "RDB::setDesc:  error with file '" + _filename + "':  " );
    throw( e );

  } catch ( ... ) {
    throw( RDBErr( "RDB::setDesc:  unexpected exception caught with file '" + _filename + "'" ) );
  }
}

/**
 * \param idx index of RDBColumn.
 * \param data user-supplied memory for storing data.
 * \param nelems number of elements in user-supplied data.
 *
 * \exception RDBErr if index is out of range.
 * \exception RDBErr if data is invalid datatype for RDBColumn.
 *
 */
void RDB::mapData( const size_t idx, double data[], const size_t nelems ) throw ( RDBErr ) {
  if ( idx > (int)_ncols ) {
    stringstream ss;
    ss << "RDB::mapData:  error with file '" << _filename << "':  index out of range [0," << idx << ")";
    throw( RDBErr( ss.str( ) ) );
  }

  try {
    _cols[idx]->mapData( data, nelems );

  } catch ( RDBErr& e ) {
    e.set_message( "RDB::mapData:  error with file '" + _filename + "':  " );
    throw( e );

  } catch ( ... ) {
    throw( RDBErr( "RDB::mapData:  unexpected exception caught with file '" + _filename + "'" ) );
  }
}

/**
 * \param idx index of RDBColumn.
 * \param data user-supplied memorgy to stored data.
 * \param nelems number of elems in user-supplied memory.
 *
 * \exception RDBErr if index is out of range.
 * \exception RDBErr if data is invalid datatype for RDBColumn.
 *
 */
void RDB::mapData( const size_t idx, long data[], const size_t nelems ) throw ( RDBErr ) {

  if ( idx > (int)_ncols ) {
    stringstream ss;
    ss << "RDB::mapData:  error with file '" << _filename << "':  index out of range [0," << idx << ")";
    throw( RDBErr( ss.str( ) ) );
  }

  try {
    _cols[idx]->mapData( data, nelems );

  } catch ( RDBErr& e ) {
    e.set_message( "RDB::mapData:  error with file '" + _filename + "':  " );
    throw( e );

  } catch ( ... ) {
    throw( RDBErr( "RDB::mapData:  unexpected exception caught with file '" + _filename + "'" ) );
  }
}

/**
 * \param idx index of RDBColumn.
 * \param data user-supplied memory to store data.
 * \param nelems number of elements in user-supplied data.
 *
 * \exception RDBErr if index is out of range.
 * \exception RDBErr if data is invalid datatype for RDBColumn.
 *
 */
void RDB::mapData( const size_t idx, string data[], const size_t nelems ) throw ( RDBErr ) {

  if ( idx > (int)_ncols ) {
    stringstream ss;
    ss << "RDB::mapData:  error with file '" << _filename << "':  index out of range [0," << idx << ")";
    throw( RDBErr( ss.str( ) ) );
  }

  try {
    _cols[idx]->mapData( data, nelems );

  } catch ( RDBErr& e ) {
    e.set_message( "RDB::mapData:  error with file '" + _filename + "':  " );
    throw( e );

  } catch ( ... ) {
    throw( RDBErr( "RDB::mapData:  unexpected exception caught with file '" + _filename + "'" ) );
  }
}

/**
 * \param idx index of RDBColumn.
 * \param data value to assign to RDBColumn data.
 *
 * \exception RDBErr if index is out of range.
 * \exception RDBErr if data is invalid for RDBColumn.
 *
 */
void RDB::setData( const size_t idx, const double data ) throw ( RDBErr ) {

  if ( idx > (int)_ncols ) {
    stringstream ss;
    ss << "RDB::setData:  error with file '" << _filename << "':  index out of range [0," << idx << ")";
    throw( RDBErr( ss.str( ) ) );
  }

  try {
    _cols[idx]->setData( data );

  } catch ( RDBErr& e ) {
    e.set_message( "RDB::setData:  error with file '" + _filename + "':  " );
    throw( e );

  } catch ( ... ) {
    throw( RDBErr( "RDB::setData:  unexpected exception caught with file '" + _filename + "'" ) );
  }
}

/**
 * \param idx index of RDBColumn.
 * \param data value to assign to RDBColumn data.
 *
 * \exception RDBErr if index is out of range.
 * \exception RDBErr if data is invalid for RDBColumn.
 *
 */
void RDB::setData( const size_t idx, const long data ) throw ( RDBErr ) {

  if ( idx > (int)_ncols ) {
    stringstream ss;
    ss << "RDB::setData:  error with file '" << _filename << "':  index out of range [0," << idx << ")";
    throw( RDBErr( ss.str( ) ) );
  }

  try {
    _cols[idx]->setData( data );

  } catch ( RDBErr& e ) {
    e.set_message( "RDB::setData:  error with file '" + _filename + "':  " );
    throw( e );

  } catch ( ... ) {
    throw( RDBErr( "RDB::setData:  unexpected exception caught with file '" + _filename + "'" ) );
  }
}

/**
 * \param idx index of RDBColumn/
 * \param data value to assign RDBColumn data.
 *
 * \exception RDBErr if index is out of range.
 * \exception RDBErr if data is invalid for RDBColumn.
 *
 */
void RDB::setData( const size_t idx, const string& data ) throw ( RDBErr ) {

  if ( idx > (int)_ncols ) {
    stringstream ss;
    ss << "RDB::setData:  error with file '" << _filename << "':  index out of range [0," << idx << ")";
    throw( RDBErr( ss.str( ) ) );
  }

  try {
    _cols[idx]->setData( data );

  } catch ( RDBErr& e ) {
    e.set_message( "RDB::setData:  error with file '" + _filename + "':  " );
    throw( e );

  } catch ( ... ) {
    throw( RDBErr( "RDB::setData:  unexpected exception caught with file '" + _filename + "'" ) );
  }
}

/**
 * \param idx index of RDBColumn.
 * \param name retrieve name of RDBColumn.
 *
 * \exception RDBErr if index is out of range.
 *
 */
void RDB::getName( const size_t idx, string& name ) const throw ( RDBErr ) {

  if ( idx > (int)_ncols ) {
    stringstream ss;
    ss << "RDB::getName:  error with file '" << _filename << "':  index out of range [0," << idx << ")";
    throw( RDBErr( ss.str( ) ) );
  }

  name = _cols[idx]->getName( );

}

/**
 * \param idx index of RDBColumn.
 * \param def retrieve definition of RDBColumn.
 *
 * \exception RDBErr if index is out of range.
 *
 */
void RDB::getDef( const size_t idx, string& def  ) throw ( RDBErr ) {

  if ( idx > (int)_ncols ) {
    stringstream ss;
    ss << "RDB::getDef:  error with file '" << _filename << "':  index out of range [0," << idx << ")";
    throw( RDBErr( ss.str( ) ) );
  }

  def = _cols[idx]->getDef( );

}

/**
 * \param idx index of RDBColumn.
 * \param width retrieve width of RDBColumn.
 *
 * \exception RDBErr if index is out of range.
 *
 */
void RDB::getWidth( const size_t idx, long& width ) const throw ( RDBErr ) {

  if ( idx > (int)_ncols ) {
    stringstream ss;
    ss << "RDB::getWidth:  error with file '" << _filename << "':  index out of range [0," << idx << ")";
    throw( RDBErr( ss.str( ) ) );
  }

  width = _cols[idx]->getWidth( );

}

/**
 * \param idx index of RDBColumn.
 * \param type retrieve type of RDBColumn.
 *
 * \exception RDBErr if index is out of range.
 *
 */
void RDB::getType( const size_t idx, RDBColumn::Type& type ) const throw ( RDBErr ) {

  if ( idx > (int)_ncols ) {
    stringstream ss;
    ss << "RDB::getType:  error with file '" << _filename << "':  index out of range [0," << idx << ")";
    throw( RDBErr( ss.str( ) ) );
  }

  type = _cols[idx]->getType( );

}

/**
 * \param idx index of RDBColumn 
 * \param just retrieve just of RDBColumn.
 *
 * \exception RDBErr if index is out of range.
 *
 */
void RDB::getJust( const size_t idx, RDBColumn::Just& just ) const throw ( RDBErr ) {

  if ( idx > (int)_ncols ) {
    stringstream ss;
    ss << "RDB::getJust:  error with file '" << _filename << "':  index out of range [0," << idx << ")";
    throw( RDBErr( ss.str( ) ) );
  }

  just = _cols[idx]->getJust( );

}

/**
 * \param idx index of RDBColumn 
 * \param desc retrieve description of RDBColumn.
 *
 * \exception RDBErr if index is out of range.
 *
 */
void RDB::getDesc( const size_t idx, string& desc ) const throw ( RDBErr ) {

  if ( idx > (int)_ncols ) {
    stringstream ss;
    ss << "RDB::getDesc:  error with file '" << _filename << "':  index out of range [0," << idx << ")";
    throw( RDBErr( ss.str( ) ) );
  }

  desc = _cols[idx]->getDesc( );

}

/**
 * \param idx index of RDBColumn 
 * \param data retrieve data of RDBColumn.
 *
 * \exception RDBErr if index is out of range.
 * \exception RDBErr if data is invalid datatype for RDBColumn.
 *
 */
void RDB::getData( const size_t idx, double& data ) throw ( RDBErr ) {

  if ( idx > (int)_ncols ) {
    stringstream ss;
    ss << "RDB::getData:  error with file '" << _filename << "':  index out of range [0," << idx << ")";
    throw( RDBErr( ss.str( ) ) );
  }

  try {
    data = _cols[idx]->getDataDouble( );

  } catch ( RDBErr& e ) {
    e.set_message( "RDB::getData:  error with file '" + _filename + "':  " );
    throw( e );

  } catch ( ... ) {
    throw( RDBErr( "RDB::getData:  unexpected exception caught with file '" + _filename + "'" ) );
  }
}

/**
 * \param idx index of RDBColumn 
 * \param data retrieve data of RDBColumn.
 *
 * \exception RDBErr if index is out of range.
 * \exception RDBErr if data is invalid datatype for RDBColumn.
 *
 */
void RDB::getData( const size_t idx, long& data ) throw ( RDBErr ) {

  if ( idx > (int)_ncols ) {
    stringstream ss;
    ss << "RDB::getData:  error with file '" << _filename << "':  index out of range [0," << idx << ")";
    throw( RDBErr( ss.str( ) ) );
  }

  try {
    data = _cols[idx]->getDataLong( );

  } catch ( RDBErr& e ) {
    e.set_message( "RDB::getData:  error with file '" + _filename + "':  " );
    throw( e );

  } catch ( ... ) {
    throw( RDBErr( "RDB::getData:  unexpected exception caught with file '" + _filename + "'" ) );
  }
}

/**
 * \param idx index of RDBColumn 
 * \param data retrieve data of RDBColumn.
 *
 * \exception RDBErr if index is out of range.
 * \exception RDBErr if data is invalid datatype for RDBColumn.
 *
 */
void RDB::getData( const size_t idx, string& data ) throw ( RDBErr ) {

  if ( idx > (int)_ncols ) {
    stringstream ss;
    ss << "RDB::getData:  error with file '" << _filename << "':  index out of range [0," << idx << ")";
    throw( RDBErr( ss.str( ) ) );
  }

  try {
    data = _cols[idx]->getDataString( );

  } catch ( RDBErr& e ) {
    e.set_message( "RDB::getData:  error with file '" + _filename + "':  " );
    throw( e );

  } catch ( ... ) {
    throw( RDBErr( "RDB::getData:  unexpected exception caught with file '" + _filename + "':  " ) );
  }
}

/**
 * \param idx index of RDBColumn 
 *
 * \exception RDBErr if index is out of range.
 *
 * \return name of RDBColumn.
 *
 */
string RDB::getName( const size_t idx ) const throw ( RDBErr ) {

  if ( idx > (int)_ncols ) {
    stringstream ss;
    ss << "RDB::getName:  error with file '" << _filename << "':  index out of range [0," << idx << ")";
    throw( RDBErr( ss.str( ) ) );
  }

  return _cols[idx]->getName( );

}

/**
 * \param idx index of RDBColumn 
 *
 * \exception RDBErr if index is out of range.
 *
 * \return definition of RDBColumn.
 *
 */
string RDB::getDef( const size_t idx  ) throw ( RDBErr ) {

  if ( idx > (int)_ncols ) {
    stringstream ss;
    ss << "RDB::getDef:  error with file '" << _filename << "':  index out of range [0," << idx << ")";
    throw( RDBErr( ss.str( ) ) );
  }

  return _cols[idx]->getDef( );

}

/**
 * \param idx index of RDBColumn 
 *
 * \exception RDBErr if index is out of range.
 *
 * \return width of RDBColumn.
 *
 */
long RDB::getWidth( const size_t idx ) const throw ( RDBErr ) {

  if ( idx > (int)_ncols ) {
    stringstream ss;
    ss << "RDB::getWidth:  error with file '" << _filename << "':  index out of range [0," << idx << ")";
    throw( RDBErr( ss.str( ) ) );
  }

  return _cols[idx]->getWidth( );

}

/**
 * \param idx index of RDBColumn 
 *
 * \exception RDBErr if index is out of range.
 *
 * \return type of RDBColumn.
 *
 */
RDBColumn::Type RDB::getType( const size_t idx ) const throw ( RDBErr ) {

  if ( idx > (int)_ncols ) {
    stringstream ss;
    ss << "RDB::getType:  error with file '" << _filename << "':  index out of range [0," << idx << ")";
    throw( RDBErr( ss.str( ) ) );
  }

  return _cols[idx]->getType( );

}

/**
 * \param idx index of RDBColumn 
 *
 * \exception RDBErr if index is out of range.
 *
 * \return justification of RDBColumn.
 *
 */
RDBColumn::Just RDB::getJust( const size_t idx ) const throw ( RDBErr ) {

  if ( idx > (int)_ncols ) {
    stringstream ss;
    ss << "RDB::getJust:  error with file '" << _filename << "':  index out of range [0," << idx << ")";
    throw( RDBErr( ss.str( ) ) );
  }

  return _cols[idx]->getJust( );

}

/**
 * \param idx index of RDBColumn 
 *
 * \exception RDBErr if index is out of range.
 *
 * \return description of RDBColumn.
 *
 */
string RDB::getDesc( const size_t idx ) const throw ( RDBErr ) {

  if ( idx > (int)_ncols ) {
    stringstream ss;
    ss << "RDB::getDesc:  error with file '" << _filename << "':  index out of range [0," << idx << ")";
    throw( RDBErr( ss.str( ) ) );
  }

  return _cols[idx]->getDesc( );

}

/**
 * \param idx index of RDBColumn 
 *
 * \exception RDBErr if index is out of range.
 * \exception RDBErr if data is invalid datatype for RDBColumn.
 *
 * \return data of RDBColumn.
 *
 */
double RDB::getDataDouble( const size_t idx ) throw ( RDBErr ) {

  if ( idx > (int)_ncols ) {
    stringstream ss;
    ss << "RDB::getDataDouble:  error with file '" << _filename << "':  index out of range [0," << idx << ")";
    throw( RDBErr( ss.str( ) ) );
  }

  try {
    return _cols[idx]->getDataDouble( );

  } catch ( RDBErr& e ) {
    e.set_message( "RDB::getDataDouble:  error with file '" + _filename + "':  " );
    throw( e );

  } catch ( ... ) {
    throw( RDBErr( "RDB::getDataDouble:  unexpected exception caught with file '" + _filename + "':" ) );
  }
}

/**
 * \param idx index of RDBColumn 
 *
 * \exception RDBErr if index is out of range.
 * \exception RDBErr if data is invalid datatype for RDBColumn.
 *
 * \return Data of RDBColumn.
 *
 */
long RDB::getDataLong( const size_t idx ) throw ( RDBErr ) {

  if ( idx > (int)_ncols ) {
    stringstream ss;
    ss << "RDB::getDataLong:  error with file '" << _filename << "':  index out of range [0," << idx << ")";
    throw( RDBErr( ss.str( ) ) );
  }

  try {
    return _cols[idx]->getDataLong( );

  } catch ( RDBErr& e ) {
    e.set_message( "RDB::getDataLong:  error with file '" + _filename + "':  " );
    throw( e );

  } catch ( ... ) {
    throw( RDBErr( "RDB::getDataLong:  unexpected exception caught with file '" + _filename + "':" ) );
  }
}

/**
 * \param idx index of RDBColumn 
 *
 * \exception RDBErr if index is out of range.
 * \exception RDBErr if data is invalid datatype for RDBColumn.
 *
 * \return data of RDBColumn.
 *
 */
string RDB::getDataString( const size_t idx ) throw ( RDBErr ) {

  if ( idx > (int)_ncols ) {
    stringstream ss;
    ss << "RDB::getDataString:  error with file '" << _filename << "':  index out of range [0," << idx << ")";
    throw( RDBErr( ss.str( ) ) );
  }

  try {
    return _cols[idx]->getDataString( );

  } catch ( RDBErr& e ) {
    e.set_message( "RDB::getDataString:  error with file '" + _filename + "':  " );
    throw( e );

  } catch ( ... ) {
    throw( RDBErr( "RDB::getDataString:  unexpected exception caught with file '" + _filename + "':" ) );
  }
}

/**
 * \param name of RDBColumn
 * \param newname new name of RDBColumn.
 *
 * \exception RDBErr if no matching column name is found.
 * \exception RDBErr if newname is invalid name for RDBColumn.
 *
 */
void RDB::setName( const string& name, const string& newname ) throw ( RDBErr ) {

  for ( size_t jdx = 0; jdx < _ncols; jdx++ ) {
    if ( name == _cols[jdx]->getName( ) ) {
      _cols[jdx]->setName( newname );
      return;

    }
  }

  if ( "_NR" != name ) {
    throw ( RDBErr( string("RDB::setName(string&,string&):  error with file '" + _filename + "':  Column ") + name + " not found" ) );
  }
  
  _nrcol.setName( newname );
  
}

/**
 * \param name of RDBColumn 
 * \param def definition of RDBColumn.
 *
 * \exception RDBErr if no matching column name is found.
 * \exception RDBErr if def is invalid definition for RDBColumn.
 *
 */
void RDB::setDef( const string& name, const string& def  ) throw ( RDBErr ) {

  for ( size_t jdx = 0; jdx < _ncols; jdx++ ) {
    if ( name == _cols[jdx]->getName( ) ) {
      try {
	_cols[jdx]->setDef( def );
	
      } catch ( RDBErr& e ) {
	e.set_message( "RDB::setDef:  error with file '" + _filename + "':  " );
	throw( e );
	
      } catch ( ... ) {
	throw( RDBErr( "RDB::setDef:  unexpected exception caught with file '" + _filename + "':" ) );
      }

      return;
      
    }
  }

  if ( "_NR" != name ) {
    throw ( RDBErr( string("RDB::setDef(string&,string&):  error with file '" + _filename + "':  Column ") + name + " not found" ) );
  }

  _nrcol.setDef( def );
}

/**
 * \param name of RDBColumn 
 * \param width width of RDBColumn.
 *
 * \exception RDBErr if no matching column name is found.
 * \exception RDBErr if width is invalid width for RDBColumn.
 *
 */
void RDB::setWidth( const string& name, const long width ) throw ( RDBErr ) {

  for ( size_t jdx = 0; jdx < _ncols; jdx++ ) {
    if ( name == _cols[jdx]->getName( ) ) {
      try {
	_cols[jdx]->setWidth( width );
	
      } catch ( RDBErr& e ) {
	e.set_message( "RDB::setWidth:  error with file '" + _filename + "':  " );
	throw( e );
	
      } catch ( ... ) {
	throw( RDBErr( "RDB::setWidth:  unexpected exception caught with file '" + _filename + "':" ) );
      }

      return;
      
    }
  }

  if ( "_NR" != name ) {
  throw ( RDBErr( string("RDB::setWidth(string&,long):  error with file '" + _filename + "':  Column ") + name + " not found" ) );
  }
  
  _nrcol.setWidth( width );
}

/**
 * \param name of RDBColumn 
 * \param type type of RDBColumn.
 *
 * \exception RDBErr if no matching column name is found.
 * \exception RDBErr if type is invalid type for RDBColumn.
 *
 */
void RDB::setType( const string& name, const RDBColumn::Type type ) throw ( RDBErr ) {

  for ( size_t jdx = 0; jdx < _ncols; jdx++ ) {
    if ( name == _cols[jdx]->getName( ) ) {
      try {
	_cols[jdx]->setType( type );
	
      } catch ( RDBErr& e ) {
	e.set_message( "RDB::setType:  error with file '" + _filename + "':  " );
	throw( e );
	
      } catch ( ... ) {
	throw( RDBErr( "RDB::setType:  error with file '" + _filename + "':  unexpected exception caught" ) );
      }

      return;

    }
  }

  if ( "_NR" != name ) {
  throw ( RDBErr( string("RDB::setType(string&,RDBColumn::Type):  error with file '" + _filename + "':  Column ") + name + " not found" ) );
  }
  
  _nrcol.setType( type );
}

/**
 * \param name of RDBColumn 
 * \param just justification of RDBColumn.
 *
 * \exception RDBErr if no matching column name is found.
 * \exception RDBErr if just is invalid justification for RDBColumn.
 *
 */
void RDB::setJust( const string& name, const RDBColumn::Just just ) throw ( RDBErr ) {

  for ( size_t jdx = 0; jdx < _ncols; jdx++ ) {
    if ( name == _cols[jdx]->getName( ) ) {
      try {
	_cols[jdx]->setJust( just );
	
      } catch ( RDBErr& e ) {
	e.set_message( "RDB::setJust:  error with file '" + _filename + "':  " );
	throw( e );
	
      } catch ( ... ) {
	throw( RDBErr( "RDB::setJust:  error with file '" + _filename + "':  unexpected exception caught" ) );
      }

      return;

    }
  }

  if ( "_NR" != name ) {
  throw ( RDBErr( string("RDB::setJust(string&,RDBColumn::Just):  error with file '" + _filename + "':  Column ") + name + " not found" ) );
  }
  
  _nrcol.setJust( just );
}

/**
 * \param name of RDBColumn 
 * \param desc description of RDBColumn.
 *
 * \exception RDBErr if no matching column name is found.
 * \exception RDBErr if desc is invalid description for RDBColumn.
 *
 */
void RDB::setDesc( const string& name, const string& desc ) throw ( RDBErr ) {

  for ( size_t jdx = 0; jdx < _ncols; jdx++ ) {
    if ( name == _cols[jdx]->getName( ) ) {
      try {
	_cols[jdx]->setDesc( desc );
	
      } catch ( RDBErr& e ) {
	e.set_message( "RDB::setDesc:  error with file '" + _filename + "':  " );
	throw( e );
	
      } catch ( ... ) {
	throw( RDBErr( "RDB::setDesc:  error with file '" + _filename + "':  unexpected exception caught" ) );
      }

      return;

    }
  }

  if ( "_NR" != name ) {
  throw ( RDBErr( string("RDB::setDesc(string&,string&):  error with file '" + _filename + "':  Column ") + name + " not found" ) );
  }
  
  _nrcol.setDesc( desc );
}

/**
 * \param name of RDBColumn 
 * \param data user-supplied memory to store RDBColumn data.
 * \param nelems number of elems in user-supplied memory.
 *
 * \exception RDBErr if no matching column name is found.
 * \exception RDBErr if data is invalid datatype for RDBColumn.
 *
 */
void RDB::mapData( const string& name, double data[], const size_t nelems ) throw ( RDBErr ) {

  for ( size_t jdx = 0; jdx < _ncols; jdx++ ) {
    if ( name == _cols[jdx]->getName( ) ) {
      try {
	_cols[jdx]->mapData( data, nelems );
	
      } catch ( RDBErr& e ) {
	e.set_message( "RDB::mapData:  error with file '" + _filename + "':  " );
	throw( e );
	
      } catch ( ... ) {
	throw( RDBErr( "RDB::mapData:  error with file '" + _filename + "':  unexpected exception caught" ) );
      }

      return;
      
    }
  }

  if ( "_NR" != name ) {
    throw ( RDBErr( string("RDB::mapData(string&,double[],size_t):  error with file '" + _filename + "':  Column ") + name + " not found" ) );
  } else {
    throw( RDBErr( "RDB::mapData('_NR',double[],size_t):  error with file '" + _filename + "':  mapping double[] to '_NR' column not allowed" ) );
  }
}

/**
 * \param name of RDBColumn 
 * \param data user-supplied memory to store RDBColumn data.
 * \param nelems number of elems in user-supplied memory.
 *
 * \exception RDBErr if no matching column name is found.
 * \exception RDBErr if data is invalid datatype for RDBColumn.
 *
 */
void RDB::mapData( const string& name, long data[], const size_t nelems ) throw ( RDBErr ) {

  for ( size_t jdx = 0; jdx < _ncols; jdx++ ) {
    if ( name == _cols[jdx]->getName( ) ) {
      try {
	_cols[jdx]->mapData( data, nelems );
	
      } catch ( RDBErr& e ) {
	e.set_message( "RDB::mapData:  error with file '" + _filename + "':  " );
	throw( e );
	
      } catch ( ... ) {
	throw( RDBErr( "RDB::mapData:  unexpected exception caught with file '" + _filename + "':" ) );

      }

      return;
      
    }
  }

  if ( "_NR" != name ) {
  throw ( RDBErr( string("RDB::mapData(string&,long[],size_t):  error with file '" + _filename + "':  Column ") + name + " not found" ) );
  }

  _nrcol.mapData( data, nelems );
}

/**
 * \param name of RDBColumn 
 * \param data user-supplied memory to store RDBColumn data.
 * \param nelems number of elems in user-supplied memory.
 *
 * \exception RDBErr if no matching column name is found.
 * \exception RDBErr if data is invalid datatype for RDBColumn.
 *
 */
void RDB::mapData( const string& name, string data[], const size_t nelems ) throw ( RDBErr ) {

  for ( size_t jdx = 0; jdx < _ncols; jdx++ ) {
    if ( name == _cols[jdx]->getName( ) ) {
      try {
	_cols[jdx]->mapData( data, nelems );
	
      } catch ( RDBErr& e ) {
	e.set_message( "RDB::mapData:  error with file '" + _filename + "':  " );
	throw( e );
	
      } catch ( ... ) {
	throw( RDBErr( "RDB::mapData:  unexpected exception caught with file '" + _filename + "':" ) );
      }

      return;

    }
  }

  if ( "_NR" != name ) {
  throw ( RDBErr( string("RDB::mapData(string&,string[],size_t):  error with file '" + _filename + "':  Column ") + name + " not found" ) );
  } else {
    throw( RDBErr( "RDB::mapData('_NR',string[],size_t):  error with file '" + _filename + "':  mapping string[] to '_NR' column not allowed" ) );
  }
}

/**
 * \param name of RDBColumn 
 * \param data value to assign RDBColumn data.
 *
 * \exception RDBErr if no matching column name is found.
 * \exception RDBErr if data is invalid datatype for RDBColumn.
 *
 */
void RDB::setData( const string& name, const double data ) throw ( RDBErr ) {

  for ( size_t jdx = 0; jdx < _ncols; jdx++ ) {
    if ( name == _cols[jdx]->getName( ) ) {
      try {
	_cols[jdx]->setData( data );
	
      } catch ( RDBErr& e ) {
	e.set_message( "RDB::setData:  error with file '" + _filename + "':  " );
	throw( e );
	
      } catch ( ... ) {
	throw( RDBErr( "RDB::setData:  unexpected exception caught with file '" + _filename + "':" ) );
      }
      return;
    }
  }

  if ( "_NR" != name ) {
  throw ( RDBErr( string("RDB::setData(string&,double):  error with file '" + _filename + "':  Column ") + name + " not found" ) );
  }
  
  _nrcol.setData( data );
}

/**
 * \param name of RDBColumn 
 * \param data value to assign RDBColumn data.
 *
 * \exception RDBErr if no matching column name is found.
 * \exception RDBErr if data is invalid datatype for RDBColumn.
 *
 */
void RDB::setData( const string& name, const long data ) throw ( RDBErr ) {

  for ( size_t jdx = 0; jdx < _ncols; jdx++ ) {
    if ( name == _cols[jdx]->getName( ) ) {
      try {
	_cols[jdx]->setData( data );
	
      } catch ( RDBErr& e ) {
	e.set_message( "RDB::setData:  error with file '" + _filename + "':  " );
	throw( e );
	
      } catch ( ... ) {
	throw( RDBErr( "RDB::setData:  unexpected exception caught with file '" + _filename + "':" ) );
      }
      return;
    }
  }

  if ( "_NR" != name ) {
  throw ( RDBErr( string("RDB::setData(string&,long):  error with file '" + _filename + "':  Column ") + name + " not found" ) );
  }
  
  _nrcol.setData( data );
}

/**
 * \param name of RDBColumn 
 * \param data value to assign RDBColumn data.
 *
 * \exception RDBErr if no matching column name is found.
 * \exception RDBErr if data is invalid datatype for RDBColumn.
 *
 */
void RDB::setData( const string& name, const string& data ) throw ( RDBErr ) {

  for ( size_t jdx = 0; jdx < _ncols; jdx++ ) {
    if ( name == _cols[jdx]->getName( ) ) {
      try {
	_cols[jdx]->setData( data );
	
      } catch ( RDBErr& e ) {
	e.set_message( "RDB::setData:  error with file '" + _filename + "':  " );
	throw( e );

      } catch ( ... ) {
	throw( RDBErr( "RDB::setData:  unexpected exception caught with file '" + _filename + "':" ) );
      }
      return;
      
    }
  }

  if ( "_NR" != name ) {
  throw ( RDBErr( string("RDB::setData(string&,string&):  error with file '" + _filename + "':  Column ") + name + " not found" ) );
  }
  
  _nrcol.setData( data );
}

/**
 * \param name of RDBColumn 
 * \param namefound name of RDBColumn if found.
 *
 * \exception RDBErr if no matching column name is found.
 *
 */
void RDB::getName( const string& name, string& namefound ) const throw ( RDBErr ) {

  for ( size_t jdx = 0; jdx < _ncols; jdx++ ) {
    if ( name == _cols[jdx]->getName( ) ) {
      namefound = _cols[jdx]->getName( );
      return;

    }
  }

  if ( "_NR" != name ) {
    throw ( RDBErr( string("RDB::getName(string&,string&):  error with file '" + _filename + "':  Column ") + name + " not found" ) );
  }
  
  namefound = _nrcol.getName( );
}

/**
 * \param name of RDBColumn 
 * \param def definition of RDBColumn.
 *
 * \exception RDBErr if no matching column name is found.
 *
 */
void RDB::getDef( const string& name, string& def  ) throw ( RDBErr ) {

  for ( size_t jdx = 0; jdx < _ncols; jdx++ ) {
    if ( name == _cols[jdx]->getName( ) ) {
      def = _cols[jdx]->getDef( );
      return;
      
    }
  }

  if ( "_NR" != name ) {
    throw ( RDBErr( string("RDB::getDef(string&,string&):  error with file '" + _filename + "':  Column ") + name + " not found" ) );
  }
  
  def = _nrcol.getDef(  );
}

/**
 * \param name of RDBColumn 
 * \param width width of RDBColumn.
 *
 * \exception RDBErr if no matching column name is found.
 *
 */
void RDB::getWidth( const string& name, long& width ) const throw ( RDBErr ) {

  for ( size_t jdx = 0; jdx < _ncols; jdx++ ) {
    if ( name == _cols[jdx]->getName( ) ) {
      width = _cols[jdx]->getWidth( );
      return;
      
    }
  }

  if ( "_NR" != name ) {
    throw ( RDBErr( string("RDB::getWidth(string&,long):  error with file '" + _filename + "':  Column ") + name + " not found" ) );
  }
  
  width = _nrcol.getWidth( );
}

/**
 * \param name of RDBColumn 
 * \param type type of RDBColumn.
 *
 * \exception RDBErr if no matching column name is found.
 *
 */
void RDB::getType( const string& name, RDBColumn::Type& type ) const throw ( RDBErr ) {

  for ( size_t jdx = 0; jdx < _ncols; jdx++ ) {
    if ( name == _cols[jdx]->getName( ) ) {
      type = _cols[jdx]->getType( );
      return;
      
    }
  }

  if ( "_NR" != name ) {
    throw ( RDBErr( string("RDB::getType(string&,RDBColumn::Type):  error with file '" + _filename + "':  Column ") + name + " not found" ) );
  }
  
  type = _nrcol.getType( );
}

/**
 * \param name of RDBColumn 
 * \param just justification of RDBColumn.
 *
 * \exception RDBErr if no matching column name is found.
 *
 */
void RDB::getJust( const string& name, RDBColumn::Just& just ) const throw ( RDBErr ) {

  for ( size_t jdx = 0; jdx < _ncols; jdx++ ) {
    if ( name == _cols[jdx]->getName( ) ) {
      just = _cols[jdx]->getJust( );
      return;

    }
  }

  if ( "_NR" != name ) {
    throw ( RDBErr( string("RDB::getJust(string&,RDBColumn::Just):  error with file '" + _filename + "':  Column ") + name + " not found" ) );
  }
  
  just = _nrcol.getJust( );
}

/**
 * \param name of RDBColumn 
 * \param desc description of RDBColumn.
 *
 * \exception RDBErr if no matching column name is found.
 *
 */
void RDB::getDesc( const string& name, string& desc ) const throw ( RDBErr ) {

  for ( size_t jdx = 0; jdx < _ncols; jdx++ ) {
    if ( name == _cols[jdx]->getName( ) ) {
      desc = _cols[jdx]->getDesc( );
      return;
      
    }
  }

  if ( "_NR" != name ) {
    throw ( RDBErr( string("RDB::getDesc(string&,string&):  error with file '" + _filename + "':  Column ") + name + " not found" ) );
  }
  
  desc = _nrcol.getDesc( );
}

/**
 * \param name of RDBColumn 
 * \param data data of RDBColumn.
 *
 * \exception RDBErr if no matching column name is found.
 * \exception RDBErr if data is invalid datatype for RDBColumn.
 *
 */
void RDB::getData( const string& name, double& data ) throw ( RDBErr ) {

  for ( size_t jdx = 0; jdx < _ncols; jdx++ ) {
    if ( name == _cols[jdx]->getName( ) ) {
      try {
	data = _cols[jdx]->getDataDouble( );
	
      } catch ( RDBErr& e ) {
	e.set_message( "RDB::getData:  error with file '" + _filename + "':  " );
	throw( e );
	
      } catch ( ... ) {
	throw( RDBErr( "RDB::getData:  unexpected exception caught with file '" + _filename + "':" ) );
      }
      return;
      
    }
  }
  
  if ( "_NR" != name ) {
    throw ( RDBErr( string("RDB::getData(string&,double):  error with file '" + _filename + "':  Column ") + name + " not found" ) );
  }
  
  data = _nrcol.getDataDouble( );
}

/**
 * \param name of RDBColumn 
 * \param data data of RDBColumn.
 *
 * \exception RDBErr if no matching column name is found.
 * \exception RDBErr if data is invalid datatype for RDBColumn.
 *
 */
void RDB::getData( const string& name, long& data ) throw ( RDBErr ) {

  for ( size_t jdx = 0; jdx < _ncols; jdx++ ) {
    if ( name == _cols[jdx]->getName( ) ) {
      try {
	data = _cols[jdx]->getDataLong( );
	
      } catch ( RDBErr& e ) {
	e.set_message( "RDB::getData:  error with file '" + _filename + "':  " );
	throw( e );
	
      } catch ( ... ) {
	throw( RDBErr( "RDB::getData:  unexpected exception caught with file '" + _filename + "':" ) );
      }
      return;
      
    }
  }
  
  if ( "_NR" != name ) {
    throw ( RDBErr( string("RDB::getData(string&,long&):  error with file '" + _filename + "':  Column ") + name + " not found" ) );
  }
  
  data = _nrcol.getDataLong( );
}

/**
 * \param name of RDBColumn 
 * \param data data of RDBColumn.
 * \exception RDBErr if data is invalid datatype for RDBColumn.
 *
 * \exception RDBErr if no matching column name is found.
 *
 */
void RDB::getData( const string& name, string& data ) throw ( RDBErr ) {

  for ( size_t jdx = 0; jdx < _ncols; jdx++ ) {
    if ( name == _cols[jdx]->getName( ) ) {
      try {
	data = _cols[jdx]->getDataString( );
	
      } catch ( RDBErr& e ) {
	e.set_message( "RDB::getData:  error with file '" + _filename + "':  " );
	throw( e );
	
      } catch ( ... ) {
	throw( RDBErr( "RDB::getData:  unexpected exception caught with file '" + _filename + "':" ) );
      }
      return;
      
    }
  }

  if ( "_NR" != name ) {
    throw ( RDBErr( string("RDB::getData(string&,string&):  error with file '" + _filename + "':  Column ") + name + " not found" ) );
  }
  
  data = _nrcol.getDataString( );
}

/**
 * \param name of RDBColumn 
 *
 * \exception RDBErr if no matching column name is found.
 *
 * \return name of RDBColumn.
 *
 */
string RDB::getName( const string& name ) const throw ( RDBErr ) {

  for ( size_t jdx = 0; jdx < _ncols; jdx++ ) {
    if ( name == _cols[jdx]->getName( ) ) {
      return _cols[jdx]->getName( );

    }
  }

  if ( "_NR" != name ) {
    throw ( RDBErr( string("RDB::getName(string&):  error with file '" + _filename + "':  Column ") + name + " not found" ) );
  }
  
  return _nrcol.getName( );
}

/**
 * \param name of RDBColumn 
 *
 * \exception RDBErr if no matching column name is found.
 *
 * \return definition of RDBColumn.
 *
 */
string RDB::getDef( const string& name  ) throw ( RDBErr ) {

  for ( size_t jdx = 0; jdx < _ncols; jdx++ ) {
    if ( name == _cols[jdx]->getName( ) ) {
      return _cols[jdx]->getDef( );

    }
  }

  if ( "_NR" != name ) {
    throw ( RDBErr( string("RDB::getDef(string&):  error with file '" + _filename + "':  Column ") + name + " not found" ) );
  }
  
  return _nrcol.getDef( );
}

/**
 * \param name of RDBColumn 
 *
 * \exception RDBErr if no matching column name is found.
 *
 * \return width of RDBColumn.
 *
 */
long RDB::getWidth( const string& name ) const throw ( RDBErr ) {

  for ( size_t jdx = 0; jdx < _ncols; jdx++ ) {
    if ( name == _cols[jdx]->getName( ) ) {
      return _cols[jdx]->getWidth( );

    }
  }

  if ( "_NR" != name ) {
    throw ( RDBErr( string("RDB::getWidth(string&):  error with file '" + _filename + "':  Column ") + name + " not found" ) );
  }
  
  return _nrcol.getWidth( );
}

/**
 * \param name of RDBColumn 
 *
 * \exception RDBErr if no matching column name is found.
 *
 * \return type of RDBColumn.
 *
 */
RDBColumn::Type RDB::getType( const string& name ) const throw ( RDBErr ) {

  for ( size_t jdx = 0; jdx < _ncols; jdx++ ) {
    if ( name == _cols[jdx]->getName( ) ) {
      return _cols[jdx]->getType( );

    }
  }

  if ( "_NR" != name ) {
    throw ( RDBErr( string("RDB::getType(string&):  error with file '" + _filename + "':  Column ") + name + " not found" ) );
  }
  
  return _nrcol.getType( );
}

/**
 * \param name of RDBColumn 
 *
 * \exception RDBErr if no matching column name is found.
 *
 * \return justification of RDBColumn.
 *
 */
RDBColumn::Just RDB::getJust( const string& name ) const throw ( RDBErr ) {

  for ( size_t jdx = 0; jdx < _ncols; jdx++ ) {
    if ( name == _cols[jdx]->getName( ) ) {
      return _cols[jdx]->getJust( );

    }
  }

  if ( "_NR" != name ) {
    throw ( RDBErr( string("RDB::getJust(string&):  error with file '" + _filename + "':  Column ") + name + " not found" ) );
  }
  
  return _nrcol.getJust( );
}

/**
 * \param name of RDBColumn 
 *
 * \exception RDBErr if no matching column name is found.
 *
 * \return description of RDBColumn.
 *
 */
string RDB::getDesc( const string& name ) const throw ( RDBErr ) {

  for ( size_t jdx = 0; jdx < _ncols; jdx++ ) {
    if ( name == _cols[jdx]->getName( ) ) {
      return _cols[jdx]->getDesc( );

    }
  }
  
  if ( "_NR" != name ) {
    throw ( RDBErr( string("RDB::getDesc(string&):  error with file '" + _filename + "':  Column ") + name + " not found" ) );
  }
  
  return _nrcol.getDesc( );
}

/**
 * \param name of RDBColumn 
 *
 * \exception RDBErr if no matching column name is found.
 * \exception RDBErr if data is invalid datatype for RDBColumn.
 *
 * \return data of RDBColumn.
 *
 */
double RDB::getDataDouble( const string& name ) throw ( RDBErr ) {

  for ( size_t jdx = 0; jdx < _ncols; jdx++ ) {
    if ( name == _cols[jdx]->getName( ) ) {
      try {
	return _cols[jdx]->getDataDouble( );
	
      } catch ( RDBErr& e ) {
	e.set_message( "RDB::getDataDouble:  error with file '" + _filename + "':  " );
	throw( e );
	
      } catch ( ... ) {
	throw( RDBErr( "RDB::getDataDouble:  unexpected exception caught with file '" + _filename + "':" ) );
      }
    }
  }

  if ( "_NR" != name ) {
  throw ( RDBErr( string("RDB::getDataDouble(string&):  error with file '" + _filename + "':  Column ") + name + " not found" ) );
  }
  
  return _nrcol.getDataDouble( );
}

/**
 * \param name of RDBColumn 
 *
 * \exception RDBErr if no matching column name is found.
 * \exception RDBErr if data is invalid datatype for RDBColumn.
 *
 * \return data of RDBColumn.
 *
 */
long RDB::getDataLong( const string& name ) throw ( RDBErr ) {

  for ( size_t jdx = 0; jdx < _ncols; jdx++ ) {
    if ( name == _cols[jdx]->getName( ) ) {
      try {
	return _cols[jdx]->getDataLong( );
	
      } catch ( RDBErr& e ) {
	e.set_message( "RDB::getDataLong:  error with file '" + _filename + "':  " );
	throw( e );
	
      } catch ( ... ) {
	throw( RDBErr( "RDB::getDataLong:  unexpected exception caught with file '" + _filename + "':" ) );
      }
    }
  }

  if ( "_NR" != name ) {
  throw ( RDBErr( string("RDB::getDataLong(string&):  error with file '" + _filename + "':  Column ") + name + " not found" ) );
  }
  
  return _nrcol.getDataLong( );
}

/**
 * \param name of RDBColumn 
 *
 * \exception RDBErr if no matching column name is found.
 * \exception RDBErr if data is invalid datatype for RDBColumn.
 *
 * \return data of RDBColumn.
 *
 */
string RDB::getDataString( const string& name ) throw ( RDBErr ) {

  for ( size_t jdx = 0; jdx < _ncols; jdx++ ) {
    if ( name == _cols[jdx]->getName( ) ) {
      try {
	return _cols[jdx]->getDataString( );
	
      } catch ( RDBErr& e ) {
	e.set_message( "RDB::getDataString:  error with file '" + _filename + "':  " );
	throw( e );
	
      } catch ( ... ) {
	throw( RDBErr( "RDB::getDataString:  unexpected exception caught with file '" + _filename + "':" ) );
      }
    }
  }

  if ( "_NR" != name ) {
    throw ( RDBErr( string("RDB::getDataString(string&):  error with file '" + _filename + "':  Column ") + name + " not found" ) );
  }
  
  return _nrcol.getDataString( );
}

/**
 * \return Number of comments.
 *
 */
size_t
RDB::nComments( 
  void
  ) const { 

  return _ncomms;

}

/**
 * \return Number of columns.
 *
 */
size_t
RDB::nColumns( 
  void
  ) const { 

  return _ncols;

}

/**
 * \return Number of rows.
 *
 * \warning This method reads through the entire RDB file to count the
 * number of rows and then attempts to rewind the file to the location
 * at which it began .  If the stream attached to the RDB object is
 * not seekable, this method will fail.
 *
 */
size_t
RDB::nRows(
  void
  ) { 

  if ( ! _knowrows ) {
    _nrows = 0;
    long position;

    if ( ios::in & _mode && _isptr ) {
      _isptr->clear( );

      position = _isptr->tellg( );
      _isptr->seekg( _rewindto );

      while ( EOF != _isptr->peek( ) && getline( *_isptr, _line, '\n' ) )
	_nrows++;
    
      _knowrows = true;
      _isptr->clear( );
      _isptr->seekg( position );

    }	
  }    

  return _nrows;

}

/**
 * Parses out comments and column names and defintions from the RDB
 * file header.  This is used by RDB#open().
 *
 */
void 
RDB::parseHeader( 
  void
  ) throw ( RDBErr ) {

  string line0, line1;
  getline( *_isptr, line0, '\n' );

  while ( suplib::iscomment( line0 ) ) {
    try { 
      RDBComment tmpcomm(line0);
      setComment( tmpcomm, -1 );

    } catch ( RDBErr& e ) {
      stringstream ss;
      ss << "RDB::parseHeader(void):  error at comment " << _ncomms + _nrows << ":  ";
      e.set_message( ss.str( ) );
      throw( e );

    } catch ( ... ) {
      stringstream ss;
      ss << "RDB::parseHeader(void):  unexpected exception caught at comment " << _ncomms + _nrows ;
      throw( RDBErr( ss.str( ) ) );

    }

    line0.erase( );
    getline( *_isptr, line0, '\n' );
  }

  getline( *_isptr, line1, '\n' );
  vector<string> nametokens;
  vector<string> deftokens; 
  suplib::tok( nametokens, line0, "\t", false );
  suplib::tok( deftokens,  line1, "\t", false );

  if ( nametokens.size( ) != deftokens.size( ) ) {
    stringstream ss;
    ss << "RDB::parseHeader(void):  Number of names(" << nametokens.size( ) << ") " << (nametokens.size() > deftokens.size() ? "greater" : "less") << " than number of definitions(" << deftokens.size( ) << ")";
    throw( RDBErr( ss.str( ) ) );

  } else {
    _ncols = nametokens.size( );

    if ( _cols ) {
      for ( size_t idx = 0; idx < _ncols; idx++ ) {
	if ( _mycols[idx] ) {
	  delete _cols[idx]; 

	}
      }
      delete [] _cols;
      delete [] _mycols;

    }

    _cols   = new RDBColumn*[_ncols];
    _mycols = new bool[_ncols];

    size_t idx;
    size_t typepos;
    for ( idx = 0; idx < _ncols; idx++ ) {
      try { 
	if ( deftokens[idx].npos != (typepos = deftokens[idx].find_first_not_of( "0123456789" )) && 'N' == deftokens[idx][typepos] ) {
	  _cols[idx] 
	    = new RDBDoubleColumn( nametokens[idx], deftokens[idx] );
	  
	} else {
	  _cols[idx] 
	    = new RDBStringColumn( nametokens[idx], deftokens[idx] );
	  
	}
      } catch ( RDBErr& rdberr ) {
	rdberr.set_message( "RDB::parseHeader(void):  column definition error:  " );
	throw( rdberr );
	
      } catch ( ... ) {
	throw( RDBErr( "RDB::parse::Header(void):  unexpected exception caught" ) );
	
      }

      _mycols[idx] = true;

    }
  }

  _rownum   = 0;
  _rewindto = _isptr->tellg( );

}

/**
 * \param line the line to be split.
 *
 * \return Vector of tokens, one for each tab delimited field.
 *
 * Splits the input line on tabs.
 *
 */
vector<string>
RDB::parseLine( 
  const string& line
  ) const {

  vector<string> tokens;
  if ( ! line.empty( ) ) {
    size_t st0 = 0, st1 = 0;

    do {
      st1 = line.find( '\t', st0 );
      if ( line.size( ) < st1 ) {
	st1 = line.size( );

      }

      tokens.push_back( line.substr( st0, st1 - st0 ) );
      st0 = st1 + 1;

    } while ( line.size( ) > st1 );
  }

  return tokens;

}

/**
 * \return number of tokens parsed.
 *
 * Splits the line on tabs and assigns the tokens to the RDBColumns.
 *
 */
size_t
RDB::parseLine(
  bool& newgroup
  ) throw ( RDBErr ) { 

  size_t ntoks = 0;
  size_t begin = 0;
  size_t end   = _line.find( '\t' );

  try {
    do {
      end = _line.find( '\t', begin );

      _cols[ntoks]->setData( _line.substr( begin, end - begin ) );

      if ( RBOG & _cols[ntoks]->newGroup( ) )
	newgroup = true;

      ntoks++;
      begin = end + 1;

    } while (  _line.npos != end );

  } catch ( RDBErr& rdberr ) {
    stringstream ss;
    ss << "RDB::parseLine(void):  error at row " << _rownum << ":  ";
    rdberr.set_message( ss.str( ) );
    throw( rdberr );

  } catch ( ... ) {
    stringstream ss;
    ss << "RDB::parseLine(void):  unexpected exception caught at row " << _rownum;
    throw( RDBErr( ss.str( ) ) );

  }

  return ntoks;

}
